/*
 * Copyright (c) 2012, Freescale Semiconductor, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "gstglvivante.h"
#include <gstbufmeta.h>

#define GL_GLEXT_PROTOTYPES
#include <GLES2/gl2ext.h>

typedef struct
{
  gfloat sampler_left;
  gfloat sampler_right;
  gfloat sampler_top;
  gfloat sampler_bottom;
} GstGLVivante;

gpointer
gst_gl_vivante_new ()
{
  GstGLVivante *viv = g_new (GstGLVivante, 1);

  viv->sampler_left = 0.0f;
  viv->sampler_right = 1.0f;
  viv->sampler_top = 1.0f;
  viv->sampler_bottom = 0.0f;

  return viv;
}

void
gst_gl_vivante_delete (gpointer gpu)
{
  g_free (gpu);
}

gboolean
gst_gl_vivante_check_format (gpointer gpu, GstVideoFormat format, gint width,
    gint height)
{
  switch (format) {
    case GST_VIDEO_FORMAT_YV12:
    case GST_VIDEO_FORMAT_NV12:
      if ((width & 15) == 0)
        return TRUE;
      break;
    default:
      GST_WARNING ("GstVideoFormat %d not supported by vivante", format);
  }
  GST_WARNING
      ("Check format %d, width,height [%d, %d] failed, can't direct render",
      format, width, height);
  return FALSE;
}

GLenum
gst_gl_vivante_get_gl_format (gpointer gpu, GstVideoFormat format)
{
  switch (format) {
    case GST_VIDEO_FORMAT_YV12:
      return GL_VIV_YV12;
    case GST_VIDEO_FORMAT_NV12:
      return GL_VIV_NV12;
    default:
      GST_WARNING ("GstVideoFormat %d not supported by vivante", format);
  }
  return GL_NONE;
}

static gboolean
format_is_yuv (GstVideoFormat format)
{
  switch (format) {
    case GST_VIDEO_FORMAT_YV12:
    case GST_VIDEO_FORMAT_NV12:
      return TRUE;
    default:
      return FALSE;
  }
}

static void
copy_data_to_gpu (gpointer gpu, GstVideoFormat format, gpointer data,
    gpointer planes[], gint width, gint height)
{
  gint src_height;
  gint src_y_size, src_uv_size;
  gint dst_y_size, dst_uv_size;

  /* width is already 16 aligned */
  switch (format) {
    case GST_VIDEO_FORMAT_YV12:
      src_height = GST_ROUND_UP_2 (height);
      src_y_size = width * src_height;
      src_uv_size = src_y_size / 4;
      dst_y_size = width * height;
      dst_uv_size = dst_y_size / 4;
      memcpy (planes[0], data, dst_y_size);
      memcpy (planes[1], (gchar *) data + src_y_size, dst_uv_size);
      memcpy (planes[2], (gchar *) data + src_y_size + src_uv_size,
          dst_uv_size);
      break;
    case GST_VIDEO_FORMAT_NV12:
      src_height = GST_ROUND_UP_2 (height);
      src_y_size = width * src_height;
      dst_y_size = width * height;
      dst_uv_size = dst_y_size / 2;
      memcpy (planes[0], data, dst_y_size);
      memcpy (planes[1], (gchar *) data + src_y_size, dst_uv_size);
      break;
    default:
      return;
  }
}

gboolean
gst_gl_vivante_tex_upload (gpointer gpu, GstGLDisplay * display)
{
  GstVideoFormat format = display->upload_video_format;
  gint width = display->upload_data_width;
  gint height = display->upload_data_height;
  gpointer data = display->upload_data;
  GstBufferMeta *meta = GST_BUFFER_META (display->upload_meta);
  gpointer planes[3];
  GLuint physical = meta ? (GLuint) (meta->physical_data) : ~0U;

  GST_INFO ("gst_gl_vivante_tex_upload physical address 0x%x, data %p, tex %d",
      physical, data, display->upload_outtex);

  if (!gst_gl_vivante_check_format (gpu, format, width, height))
    return FALSE;

  glBindTexture (GL_TEXTURE_2D, display->upload_outtex);

  if (((guint) data & 0x3F) != 0) {     /* not aligned, need additional copy */
    if (!format_is_yuv (format)) {      /* rgb is left to default handler */
      GST_WARNING ("Format %d unaligned data is not supported for"
          " direct rendering", format);
      return FALSE;
    }

    GST_WARNING ("Data not aligned, need additional copy");

    glTexDirectVIV (GL_TEXTURE_2D, width, height,
        gst_gl_vivante_get_gl_format (gpu, format), (GLvoid **) & planes);
    if (glGetError () != GL_NO_ERROR)
      return FALSE;

    copy_data_to_gpu (gpu, format, data, planes, width, height);
  } else {
    glTexDirectVIVMap (GL_TEXTURE_2D, width, height,
        gst_gl_vivante_get_gl_format (gpu, format), &data, &physical);
    if (glGetError () != GL_NO_ERROR)
      return FALSE;
    display->upload_buf_mapped = TRUE;
  }

  glTexDirectInvalidateVIV (GL_TEXTURE_2D);

  return TRUE;
}

void
gst_gl_vivante_set_caps (gpointer gpu, GstCaps * caps)
{
  GstVideoFormat format;
  gint width;
  gint height;
  GstStructure *s;
  gint left = 0, right = 0, top = 0, bottom = 0;
  GstGLVivante *viv = (GstGLVivante *) gpu;

  s = gst_caps_get_structure (caps, 0);
  gst_structure_get_int (s, "crop-left", &left);
  gst_structure_get_int (s, "crop-top", &top);
  gst_structure_get_int (s, "crop-right", &right);
  gst_structure_get_int (s, "crop-bottom", &bottom);

  gst_video_format_parse_caps (caps, &format, &width, &height);

  viv->sampler_left = ((gfloat) left) / width;
  viv->sampler_right = ((gfloat) (width - right)) / width;
  viv->sampler_top = ((gfloat) (height - bottom)) / height;
  viv->sampler_bottom = ((gfloat) top) / height;
}

void
gst_gl_vivante_get_sampler_rect (gpointer gpu, gfloat * left, gfloat * right,
    gfloat * top, gfloat * bottom)
{
  GstGLVivante *viv = (GstGLVivante *) gpu;

  *left = viv->sampler_left;
  *right = viv->sampler_right;
  *top = viv->sampler_top;
  *bottom = viv->sampler_bottom;
}
