/*
 * GStreamer
 * Copyright (C) 2008 Julien Isorce <julien.isorce@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstglwindow.h"
#include "EGL/eglvivante.h"

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#define GST_GL_WINDOW_GET_PRIVATE(o)  \
  (G_TYPE_INSTANCE_GET_PRIVATE((o), GST_GL_TYPE_WINDOW, GstGLWindowPrivate))

/* A gl window is created and deleted in a thread dedicated to opengl calls
   The name contains "window" because an opengl context is used in cooperation
   with a window */

const gchar *EGLErrorString ();

enum
{
  ARG_0,
  ARG_DISPLAY
};

enum GstGLFbEventType
{
  FB_EVENT_DRAW,
  FB_EVENT_CALLBACK,
  FB_EVENT_QUIT,
};

typedef struct _GstGLFbEvent
{
  enum GstGLFbEventType type;
  GstGLWindowCB callback;
  gpointer data;
} GstGLFbEvent;

struct _GstGLWindowPrivate
{
  /* X is not thread safe */
  GMutex *lock;
  GCond *cond_send_message;
  GCond *cond_queue_message;
  gboolean running;
  gboolean allow_extra_expose_events;

  /* fb queue */
  GList *queue;

  /* X context */
  gchar *display_name;
  gint device_width;
  gint device_height;

  EGLNativeWindowType internal_win_id;
  EGLNativeDisplayType device;

  /* EGL */
  EGLContext gl_context;
  EGLDisplay gl_display;
  EGLSurface gl_surface;

  /* frozen callbacks */
  GstGLWindowCB draw_cb;
  gpointer draw_data;
  GstGLWindowCB2 resize_cb;
  gpointer resize_data;
  GstGLWindowCB close_cb;
  gpointer close_data;
};

G_DEFINE_TYPE (GstGLWindow, gst_gl_window, G_TYPE_OBJECT);

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "GstGLWindow"

gboolean _gst_gl_window_debug = FALSE;

void
gst_gl_window_init_platform ()
{
}

static void
free_event (gpointer event, gpointer data)
{
  g_slice_free (GstGLFbEvent, event);
}

/* Must be called in the gl thread */
static void
gst_gl_window_finalize (GObject * object)
{
  GstGLWindow *window = GST_GL_WINDOW (object);
  GstGLWindowPrivate *priv = window->priv;
  gboolean ret = TRUE;

  g_mutex_lock (priv->lock);

  g_debug ("about to finalize gl window\n");

  if (priv->gl_context) {
    ret =
        eglMakeCurrent (priv->gl_display, EGL_NO_SURFACE, EGL_NO_SURFACE,
        EGL_NO_CONTEXT);
    if (!ret)
      g_debug ("failed to release opengl context\n");

    eglDestroyContext (priv->gl_display, priv->gl_context);
  }

  if (priv->gl_display)
    eglTerminate (priv->gl_display);

  if (priv->device)
    fbDestroyDisplay (priv->device);

  g_list_foreach (priv->queue, free_event, NULL);
  g_list_free (priv->queue);
  priv->queue = NULL;

  if (priv->cond_send_message) {
    g_cond_free (priv->cond_send_message);
    priv->cond_send_message = NULL;
  }

  if (priv->cond_queue_message) {
    g_cond_free (priv->cond_queue_message);
    priv->cond_queue_message = NULL;
  }

  g_mutex_unlock (priv->lock);

  if (priv->lock) {
    g_mutex_free (priv->lock);
    priv->lock = NULL;
  }

  G_OBJECT_CLASS (gst_gl_window_parent_class)->finalize (object);
}

static void
gst_gl_window_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstGLWindow *window;
  GstGLWindowPrivate *priv;

  g_return_if_fail (GST_GL_IS_WINDOW (object));

  window = GST_GL_WINDOW (object);

  priv = window->priv;

  switch (prop_id) {
    case ARG_DISPLAY:
      priv->display_name = g_strdup (g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_gl_window_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstGLWindow *window;
  GstGLWindowPrivate *priv;

  g_return_if_fail (GST_GL_IS_WINDOW (object));

  window = GST_GL_WINDOW (object);

  priv = window->priv;

  switch (prop_id) {
    case ARG_DISPLAY:
      g_value_set_string (value, priv->display_name);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_gl_window_log_handler (const gchar * domain, GLogLevelFlags flags,
    const gchar * message, gpointer user_data)
{
  if (_gst_gl_window_debug) {
    g_log_default_handler (domain, flags, message, user_data);
  }
}

static void
gst_gl_window_class_init (GstGLWindowClass * klass)
{
  GObjectClass *obj_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GstGLWindowPrivate));

  obj_class->finalize = gst_gl_window_finalize;
  obj_class->set_property = gst_gl_window_set_property;
  obj_class->get_property = gst_gl_window_get_property;

  g_object_class_install_property (obj_class, ARG_DISPLAY,
      g_param_spec_string ("display", "Display", "X Display name", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

static void
gst_gl_window_init (GstGLWindow * window)
{
  GstGLWindowPrivate *priv = NULL;
  window->priv = GST_GL_WINDOW_GET_PRIVATE (window);
  priv = window->priv;

  if (g_getenv ("GST_GL_WINDOW_DEBUG") != NULL)
    _gst_gl_window_debug = TRUE;

  g_log_set_handler ("GstGLWindow", G_LOG_LEVEL_DEBUG,
      gst_gl_window_log_handler, NULL);

  priv->lock = NULL;
  priv->cond_send_message = NULL;
  priv->running = FALSE;
  priv->allow_extra_expose_events = FALSE;

  priv->queue = NULL;

  /* X context */
  priv->display_name = NULL;
  priv->device_width = 0;
  priv->device_height = 0;

  /* EGL */
  priv->gl_context = EGL_NO_CONTEXT;
  priv->gl_display = 0;
  priv->gl_surface = EGL_NO_SURFACE;

  /* frozen callbacks */
  priv->draw_cb = NULL;
  priv->draw_data = NULL;
  priv->resize_cb = NULL;
  priv->resize_data = NULL;
  priv->close_cb = NULL;
  priv->close_data = NULL;
}

/* Must be called in the gl thread */
GstGLWindow *
gst_gl_window_new (gulong external_gl_context)
{
  GstGLWindow *window = g_object_new (GST_GL_TYPE_WINDOW, NULL);
  GstGLWindowPrivate *priv = window->priv;

  EGLint config_attrib[] = {
    EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
    EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
    EGL_DEPTH_SIZE, 16,
    EGL_NONE
  };

  EGLint context_attrib[] = {
    EGL_CONTEXT_CLIENT_VERSION, 2,
    EGL_NONE
  };

  EGLint majorVersion;
  EGLint minorVersion;
  EGLint numConfigs;
  EGLConfig config;

  int index = 0;

  setlocale (LC_NUMERIC, "C");

  priv->lock = g_mutex_new ();
  priv->cond_send_message = g_cond_new ();
  priv->cond_queue_message = g_cond_new ();
  priv->running = TRUE;
  priv->allow_extra_expose_events = TRUE;

  g_mutex_lock (priv->lock);

  if (priv->display_name)
    index = strtol (priv->display_name, NULL, 0);

  priv->device = fbGetDisplayByIndex (index);

  if (!priv->device) {
    g_debug ("failed to get fb display by index %d\n", index);
    goto failure;
  }

  priv->internal_win_id = fbCreateWindow (priv->device, -1, -1, 0, 0);

  fbGetDisplayGeometry (priv->device, &priv->device_width,
      &priv->device_height);

  priv->gl_display = eglGetDisplay (priv->device);

  if (eglInitialize (priv->gl_display, &majorVersion, &minorVersion))
    g_debug ("egl initialized: %d.%d\n", majorVersion, minorVersion);
  else {
    g_debug ("failed to initialize egl %ld, %s\n", (gulong) priv->gl_display,
        EGLErrorString ());
    goto failure;
  }

  if (eglChooseConfig (priv->gl_display, config_attrib, &config, 1,
          &numConfigs))
    g_debug ("config set: %ld, %ld\n", (gulong) config, (gulong) numConfigs);
  else {
    g_debug ("failed to set config %ld, %s\n", (gulong) priv->gl_display,
        EGLErrorString ());
    goto failure;
  }

  priv->gl_surface =
      eglCreateWindowSurface (priv->gl_display, config, priv->internal_win_id,
      NULL);
  if (priv->gl_surface != EGL_NO_SURFACE)
    g_debug ("surface created: %ld\n", (gulong) priv->gl_surface);
  else {
    g_debug ("failed to create surface %ld, %ld, %ld, %s\n",
        (gulong) priv->gl_display, (gulong) priv->gl_surface,
        (gulong) priv->gl_display, EGLErrorString ());
    goto failure;
  }

  g_debug ("about to create gl context\n");

  priv->gl_context =
      eglCreateContext (priv->gl_display, config,
      (EGLContext) (guint) external_gl_context, context_attrib);

  if (priv->gl_context != EGL_NO_CONTEXT)
    g_debug ("gl context created: %ld\n", (gulong) priv->gl_context);
  else {
    g_debug ("failed to create glcontext %ld, %ld, %s\n",
        (gulong) priv->gl_context, (gulong) priv->gl_display,
        EGLErrorString ());
    goto failure;
  }

  if (!eglMakeCurrent (priv->gl_display, priv->gl_surface, priv->gl_surface,
          priv->gl_context)) {
    g_debug ("failed to make opengl context current %ld, %s\n",
        (gulong) priv->gl_display, EGLErrorString ());
    goto failure;
  }

  g_mutex_unlock (priv->lock);
  printf ("device WxH:%dx%d\n", priv->device_width, priv->device_height);
  return window;

failure:
  g_mutex_unlock (priv->lock);
  g_object_unref (G_OBJECT (window));
  return NULL;
}

GQuark
gst_gl_window_error_quark (void)
{
  return g_quark_from_static_string ("gst-gl-window-error");
}

gulong
gst_gl_window_get_internal_gl_context (GstGLWindow * window)
{
  GstGLWindowPrivate *priv = window->priv;
  return (gulong) priv->gl_context;
}

void
callback_activate_gl_context (GstGLWindowPrivate * priv)
{
  if (!eglMakeCurrent (priv->gl_display, priv->gl_surface, priv->gl_surface,
          priv->gl_context))
    g_debug ("failed to activate opengl context %lud\n",
        (gulong) priv->gl_context);
}

void
callback_inactivate_gl_context (GstGLWindowPrivate * priv)
{
  if (!eglMakeCurrent (priv->device, EGL_NO_SURFACE, EGL_NO_SURFACE,
          EGL_NO_CONTEXT))
    g_debug ("failed to inactivate opengl context %lud\n",
        (gulong) priv->gl_context);
}

void
gst_gl_window_activate_gl_context (GstGLWindow * window, gboolean activate)
{
  GstGLWindowPrivate *priv = window->priv;
  if (activate)
    gst_gl_window_send_message (window,
        GST_GL_WINDOW_CB (callback_activate_gl_context), priv);
  else
    gst_gl_window_send_message (window,
        GST_GL_WINDOW_CB (callback_inactivate_gl_context), priv);
}

/* Not called by the gl thread */
void
gst_gl_window_set_external_window_id (GstGLWindow * window, gulong id)
{
}

void
gst_gl_window_set_draw_callback (GstGLWindow * window, GstGLWindowCB callback,
    gpointer data)
{
  GstGLWindowPrivate *priv = window->priv;

  g_mutex_lock (priv->lock);

  priv->draw_cb = callback;
  priv->draw_data = data;

  g_mutex_unlock (priv->lock);
}

void
gst_gl_window_set_resize_callback (GstGLWindow * window,
    GstGLWindowCB2 callback, gpointer data)
{
  GstGLWindowPrivate *priv = window->priv;

  g_mutex_lock (priv->lock);

  priv->resize_cb = callback;
  priv->resize_data = data;

  g_mutex_unlock (priv->lock);
}

void
gst_gl_window_set_close_callback (GstGLWindow * window, GstGLWindowCB callback,
    gpointer data)
{
  GstGLWindowPrivate *priv = window->priv;

  g_mutex_lock (priv->lock);

  priv->close_cb = callback;
  priv->close_data = data;

  g_mutex_unlock (priv->lock);
}

/* Called in the gl thread */
void
gst_gl_window_draw_unlocked (GstGLWindow * window, gint width, gint height)
{
  GstGLWindowPrivate *priv = window->priv;

  if (priv->running && priv->allow_extra_expose_events) {
    GstGLFbEvent *event = g_slice_new0 (GstGLFbEvent);
    event->type = FB_EVENT_DRAW;
    priv->queue = g_list_append (priv->queue, event);
    g_cond_signal (priv->cond_queue_message);
    /* block until opengl calls have been executed in the gl thread */
    g_cond_wait (priv->cond_send_message, priv->lock);
  }
}

/* Not called by the gl thread */
void
gst_gl_window_draw (GstGLWindow * window, gint width, gint height)
{
  if (window) {
    GstGLWindowPrivate *priv = window->priv;

    g_mutex_lock (priv->lock);

    if (priv->running) {
      GstGLFbEvent *event = g_slice_new0 (GstGLFbEvent);
      event->type = FB_EVENT_DRAW;
      priv->queue = g_list_append (priv->queue, event);
      g_cond_signal (priv->cond_queue_message);
      /* block until opengl calls have been executed in the gl thread */
      g_cond_wait (priv->cond_send_message, priv->lock);
    }

    g_mutex_unlock (priv->lock);
  }
}

static void
gst_gl_fb_queue_flush (GList * queue, GCond * cond)
{
  GstGLFbEvent *event;
  GList *first;

  while ((first = g_list_first (queue))) {
    event = first->data;
    queue = g_list_delete_link (queue, first);

    if (event->type == FB_EVENT_CALLBACK) {
      g_debug ("execute last pending custom x events\n");

      if (!event->callback || !event->data)
        g_debug ("custom cb not initialized\n");

      event->callback (event->data);

      g_cond_signal (cond);
    }

    g_slice_free (GstGLFbEvent, event);
  }
}

/* Called in the gl thread */
void
gst_gl_window_run_loop (GstGLWindow * window)
{
  GstGLWindowPrivate *priv = window->priv;
  GstGLFbEvent *event;
  GList *first;

  g_debug ("begin loop\n");

  g_mutex_lock (priv->lock);

  while (priv->running) {

    if (!priv->queue)
      g_cond_wait (priv->cond_queue_message, priv->lock);

    first = g_list_first (priv->queue);
    event = first->data;
    priv->queue = g_list_delete_link (priv->queue, first);

    // use in generic/cube and other related uses
    priv->allow_extra_expose_events = g_list_length (priv->queue) <= 2;

    switch (event->type) {
      case FB_EVENT_CALLBACK:
      {
        if (priv->running) {
          if (!event->callback || !event->data)
            g_debug ("custom cb not initialized\n");
          event->callback (event->data);
          g_cond_signal (priv->cond_send_message);
        }
        break;
      }

      case FB_EVENT_QUIT:
      {
        g_debug ("Quit loop message %lud\n", (gulong) priv->internal_win_id);

        /* exit loop */
        priv->running = FALSE;

        /* make sure last pendings send message calls are executed */
        gst_gl_fb_queue_flush (priv->queue, priv->cond_send_message);
        priv->queue = NULL;

        /* Finally we can destroy opengl ressources (texture/shaders/fbo) */
        if (!event->callback || !event->data)
          g_debug ("destroy cb not correclty set\n");

        event->callback (event->data);
        g_cond_signal (priv->cond_send_message);
        break;
      }

      case FB_EVENT_DRAW:
        if (priv->draw_cb) {
          priv->draw_cb (priv->draw_data);
          glFlush ();
          eglSwapBuffers (priv->gl_display, priv->gl_surface);
          g_cond_signal (priv->cond_send_message);
        }
        break;

      default:
        g_debug ("unknown GstGLFbEvent type: %ud\n", event->type);
        break;
    }                           // switch

    g_slice_free (GstGLFbEvent, event);

  }                             // while running

  g_mutex_unlock (priv->lock);

  g_debug ("end loop\n");
}

/* Not called by the gl thread */
void
gst_gl_window_quit_loop (GstGLWindow * window, GstGLWindowCB callback,
    gpointer data)
{
  if (window) {
    GstGLWindowPrivate *priv = window->priv;

    g_mutex_lock (priv->lock);

    if (priv->running) {
      GstGLFbEvent *event = g_slice_new0 (GstGLFbEvent);
      event->type = FB_EVENT_QUIT;
      event->callback = callback;
      event->data = data;
      priv->queue = g_list_append (priv->queue, event);

      g_cond_signal (priv->cond_queue_message);
      g_cond_wait (priv->cond_send_message, priv->lock);
    }

    g_mutex_unlock (priv->lock);
  }
}

/* Not called by the gl thread */
void
gst_gl_window_send_message (GstGLWindow * window, GstGLWindowCB callback,
    gpointer data)
{
  if (window) {
    GstGLWindowPrivate *priv = window->priv;

    g_mutex_lock (priv->lock);

    if (priv->running) {
      GstGLFbEvent *event = g_slice_new0 (GstGLFbEvent);
      event->type = FB_EVENT_CALLBACK;
      event->callback = callback;
      event->data = data;
      priv->queue = g_list_append (priv->queue, event);
      g_cond_signal (priv->cond_queue_message);

      /* block until opengl calls have been executed in the gl thread */
      g_cond_wait (priv->cond_send_message, priv->lock);
    }

    g_mutex_unlock (priv->lock);
  }
}

const gchar *
EGLErrorString ()
{
  EGLint nErr = eglGetError ();
  switch (nErr) {
    case EGL_SUCCESS:
      return "EGL_SUCCESS";
    case EGL_BAD_DISPLAY:
      return "EGL_BAD_DISPLAY";
    case EGL_NOT_INITIALIZED:
      return "EGL_NOT_INITIALIZED";
    case EGL_BAD_ACCESS:
      return "EGL_BAD_ACCESS";
    case EGL_BAD_ALLOC:
      return "EGL_BAD_ALLOC";
    case EGL_BAD_ATTRIBUTE:
      return "EGL_BAD_ATTRIBUTE";
    case EGL_BAD_CONFIG:
      return "EGL_BAD_CONFIG";
    case EGL_BAD_CONTEXT:
      return "EGL_BAD_CONTEXT";
    case EGL_BAD_CURRENT_SURFACE:
      return "EGL_BAD_CURRENT_SURFACE";
    case EGL_BAD_MATCH:
      return "EGL_BAD_MATCH";
    case EGL_BAD_NATIVE_PIXMAP:
      return "EGL_BAD_NATIVE_PIXMAP";
    case EGL_BAD_NATIVE_WINDOW:
      return "EGL_BAD_NATIVE_WINDOW";
    case EGL_BAD_PARAMETER:
      return "EGL_BAD_PARAMETER";
    case EGL_BAD_SURFACE:
      return "EGL_BAD_SURFACE";
    default:
      return "unknown";
  }
}
